from pasteob.base import *
from pasteob.errors import *
from mext.uri import Uri

__all__ = ['PathMatch', 'PathMap', 'HostMap', 'Autodispatch']


class PathMatch(object):
    """
        Dispatch by exact match

            app1 = PathMatch(root_app, view=view_app)            #/prefix* <- root_app
            app2 = PathMatch({'': root_app, '/view': view_app})  #/prefix
            app3 = PathMatch({'/': root_app, '/view': view_app}) #/prefix/

    """
    fallback = None
    def __init__(self, *args, **kw):
        self.map = {}
        for k,v in kw.items():
            self.map[u'/' + k] = v
        if args:
            arg0, = args
            if isinstance(arg0, (list, tuple, dict)):
                for path, app in dict(arg0).items():
                    self[path] = app
            else:
                self.fallback = arg0

    def __setitem__(self, path, app):
        if isinstance(path, str):
            path = path.decode('utf8')
        if not isinstance(path, unicode) or path and path[0] != '/':
            raise ValueError("Invalid PathMatch key:", path)
        if path in self.map:
            raise KeyError("Duplicate PathMatch key:", path)
        self.map[path] = app

    @wapp
    def __call__(self, req):
        key = req.upath_info
        if key not in self.map:
            return self.fallback
        req.uscript_name += key
        req.path_info = ''
        return self.map[key]





def split_path(path):
    path = path.strip('/')
    if not path:
        return () # otherwise we'd get ['']
    return path.split('/')

class PathMap(object):
    """
        Dispatch by prefix
    """
    def __init__(self, *args, **kw):
        self.index = {} # segment -> another index
                        # None -> app for current prefix match
        if args:
            arg0, = args
            # using `type() is` check because some WSGI apps are
            # in fact dict subclasses -- they should be used as is
            if type(arg0) in (list, tuple, dict):
                arg0 = dict(arg0)
                collisions = set(arg0).intersection(kw)
                if collisions:
                    raise ValueError("Key collisions:", list(collisions))
                kw.update(arg0)
            else:
                self.index[None] = arg0
        for k,v in kw.items():
            self[k] = v


    def __setitem__(self, path, app):
        assert isinstance(path, basestring)
        path = unicode(path)
        index = self.index
        for seg in split_path(path):
            index = index.setdefault(seg, {})
        if None in index:
            raise KeyError(
                "PathMap already has an app registered for this prefix: %r"
                % index[None]
            )
        index[None] = app

    def __getitem__(self, path):
        assert isinstance(path, basestring), `path`
        path = unicode(path)
        index = self.index
        app = index.get(None)
        for seg in split_path(path):
            index = index.get(seg)
            if index is None:
                break
            app = index.get(None, app)
        if app is None:
            raise KeyError(path)
        return app

    @wapp
    def __call__(self, req):
        path = req.upath_info
        index = self.index
        app = index.get(None)
        app_path = u''
        while path:
            if '/' in path:
                head, path = path.split('/', 1)
                if not head:
                    continue
            else:
                head, path = path, u''
            index = index.get(head)
            if index is None:
                break
            elif None in index:
                app = index[None]
                if path:
                    app_path = req.upath_info[:-len(path)].rstrip('/')
                else:
                    app_path = req.upath_info
        app_path = app_path.rstrip(u'/')
        req.uscript_name += app_path
        req.upath_info = req.upath_info[len(app_path):]
        return app

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        else:
            return (other.index == self.index)

    def __ne__(self, other):
        return not (self == other)



class HostMap(dict):
    fallback = None
    @wapp
    def __call__(self, req):
        return self.get(req.host.split(':')[0], self.fallback)





#pm = PathMap((('foo',1),), foo=1)
#print pm['foo'] # ValueError


class Autodispatch(object):
    """
        WSGI app that automatically dispatches requests to apps stored
        as attributes. (Usually wapp-wrapped methods).

        For the following examples we assume that an instance of Autodispatch
        subclass is mounted at '/app' path, that is, by the time request gets to
        it, script_name == '/app' and the rest of the path is in path_info.

        We give examples for GET requests, but the same dispatch mechanisms are
        applied for every other methods, so if a GET_foo is mentioned, it would
        be POST_foo if the req.method is 'POST' etc.

        HEAD requests are dispatched as if they were GET requests.

        For the purposes of this docstring, the meaning of "available" is
        "an attribute that is present and is not None".

        GET /app
            If self.match_GET is available, it will be used to generate a
            response.

            Otherwise, if self.GET is available, a temporary redirect will be
            issued to /app/ (query string is preserved, so '/app?foo=x' will
            redirect to '/app/?foo=x')

        GET /app/
            If self.GET is available it will be used to generate a response.

        GET /app/foo
            If self.GET_foo is available it will be used for response with
            script_name == '/app/foo' and path_info == ''.

            Otherwise, if self.ANY_foo is available, it will used with the
            same script_name / path_info adjustments. This is useful if you
            want some app to be mounted for a certain path and that app can
            handle all HTTP methods.

            If that is not available either, the dispatch_foo rule below
            will be used.

            There's an exception to this: if the request uses some method
            other than GET and neither `METHOD`_foo nor ANY_foo is available,
            but GET_foo is, the dispatch_foo will not be used. Instead, the
            405 Method Not Allowed response will be generated with "Allow: GET"
            header.

        GET /app/foo/
        GET /app/foo/bar
            If self.dispatch_foo is available it will be used to generate the
            response after adjusting script_name and path_info so that
            script_name == '/app/foo'

        Whenever no rule above matches, self.__default__ will be used as a fallback
        for the response. Unless it is overridden in a subclass, that results in
        404 Not Found error.
    """
    __default__ = None

    def _get_app(self, *args):
        return getattr(self, '_'.join(args), None)


    @wapp
    def __call__(self, req):
        method = req.method
        if method == 'HEAD':
            method = 'GET'
        if req.path_info.strip('/'):
            return self._call_seg(req, method)
        else:
            return self._call_noseg(req, method)

    def _call_noseg(self, req, method):
        app = None
        if req.path_info == '':
            app = self._get_app('match', method) # match_GET
            if app is None and self._get_app('match', 'GET') is not None:
                # there's no match_METH but there is match_GET,
                # so req.method is wrong
                return ErrorResponse(405, allow=['GET'])
        if app is None:
            app = self._get_app(method) # GET
            if app is not None and req.path_info == '': # redirect /app -> /app/
                uri_as_dir = Uri(req.url).as_dir(True)
                return RedirectResponse('temp', uri_as_dir)
        if app is not None:
            return app
        elif self._get_app('GET') is not None:
            # there's no .METH but there is .GET
            return ErrorResponse(405, allow=['GET'])
        else:
            return self.__default__

    def _call_seg(self, req, method):
        app = None
        seg = req.path_info_peek()
        if req.path_info == '/' + seg:
            app = self._get_app(method, seg) # GET_foo
            if app is None:
                app = self._get_app('ANY', seg) # ANY_foo
                if app is None and self._get_app('GET', seg) is not None:
                    # there's no METH_seg and no ANY_seg
                    # but there is GET_seg, so req.method is wrong
                    return ErrorResponse(405, allow=['GET'])
        if app is None:
            app = self._get_app('dispatch', seg) # dispatch_foo
        if app is not None:
            req.path_info_pop()
            return app
        else:
            return self.__default__


#@@ query for all methods

